#include <iostream>
#include <vector>
#include <thread>
#include <mutex>
#include <condition_variable>

using namespace std;
using namespace chrono;

// Propusnica za ispis na ekran:
mutex term_m;

// Tip operacija koje procesi izvrsavaju:
enum TipOperacije {
    NORMALNA,
    PREPUSTI,
    POSALJI,
    PRIMI
};

// Pojedinacna operacija sa dodatim parametrima operacije:
struct Operacija {
    TipOperacije tip;
    string sadrzaj; // Sadrzaj koji se salje u sanduce (postavlja se samo kod operacije POSALJI)
    Operacija(TipOperacije p_tip, string p_sadrzaj = "") :
        tip(p_tip), sadrzaj(p_sadrzaj) {}
};

class OS {
public:
    // id - ID procesa cije operacije se izvrsavaju
    // oper - Vektor operacija koje se izvrsavaju
    void izvrsiOperacije(int id, vector<Operacija>& oper) {
        // TODO Implementirati ovu metodu
    }
};

struct Proces {
    int pauza;
    vector<Operacija> operacije;
};

void fn(OS& os, int id, Proces& proces) {
    {
        unique_lock<mutex> lt(term_m);
        cout << "[proces][" << id << "] je pokrenut, pauzira: " << proces.pauza << " milisekundi"
             << endl;
    }
    this_thread::sleep_for(milliseconds(proces.pauza));
    os.izvrsiOperacije(id, proces.operacije);
}

// Pomocna funkcija koja sluzi za pokretanje test slucajeva:
void izvrsiProcese(OS& os, vector<Proces> procesi) {
    int br = procesi.size();
    thread t[br];
    for (int i = 0; i < br; i++)
        t[i] = thread(fn, ref(os), i, ref(procesi[i]));

    for (int i = 0; i < br; i++)
        t[i].join();
}

// ***** TEST SLUCAJEVI *****

void testPrekljuciNaKrajuIzvrsenja() {
    cout << "TEST: prekljucivanje nakon sto je proces zavrsen" << endl;
    cout << "********************" << endl;
    OS os;

    vector<Proces> procesi(2);

    // Osnovni test izvrsavanja procesa. Ovaj test proverava da li dodje do prekljucivanja nakon
    // sto se prvi proces izvrsi do kraja.
    // Kreiraju se dva procesa, primetiti da proces 1 startuje 100 ms nakon procesa 0.
    // Time se postize da uvek prvo krene izvrsenje procesa 0.
    procesi[0].operacije = {NORMALNA, NORMALNA};
    procesi[0].pauza = 0;
    procesi[1].operacije = {NORMALNA, NORMALNA};
    procesi[1].pauza = 100;

    izvrsiProcese(os, procesi);
    cout << endl;
}

void testPrekljuciNaOperacijiPrepusti() {
    cout << "TEST: prekljucivanje prilikom izvrsenja operacije PREPUSTI" << endl;
    cout << "********************" << endl;
    OS os;

    vector<Proces> procesi(2);

    // Osnovni test izvrsavanje operacije PREPUSTI. Proverava se da li dolazi do prekljucivanja
    // kada proces izvrsi operaciju PREPUSTI.
    // Kreiraju se dva procesa, primetiti da proces 1 startuje 100 ms nakon procesa 0.
    // Time se postize da uvek prvo krene izvrsenje procesa 0.
    procesi[0].operacije = {NORMALNA, PREPUSTI, NORMALNA};
    procesi[0].pauza = 0;
    procesi[1].operacije = {NORMALNA, NORMALNA, PREPUSTI};
    procesi[1].pauza = 100;

    izvrsiProcese(os, procesi);
    cout << endl;
}

void testRazmenaPorukeIzmedjuDvaProcesaBezCekanja() {
    cout << "TEST: razmena poruke izmedju 2 procesa, proces koji prima poruku nalazi spremnu "
         << "poruku u sanducetu" << endl;
    cout << "********************" << endl;
    OS os;

    vector<Proces> procesi(2);

    // Razmena poruke izmedju dva procesa. Proces 0 prvo salje poruku, koju potom proces 1 prima
    // i ispisuje je na ekran.
    procesi[0].operacije = {Operacija(POSALJI, "Zdravo"), PREPUSTI};
    procesi[0].pauza = 0;
    procesi[1].operacije = {PRIMI};
    procesi[1].pauza = 100;

    izvrsiProcese(os, procesi);
    cout << endl;
}

void testRazmenaPorukeIzmedjuTriProcesaBezCekanja() {
    cout << "TEST: razmena poruke izmedju 3 procesa, proces koji prima poruku nalazi spremnu "
         << "poruku u sanducetu" << endl;
    cout << "********************" << endl;
    OS os;

    vector<Proces> procesi(3);

    // Razmena poruke izmedju tri procesa. Proces 0 prvo postavlja svoju poruku u sanduce, posle
    // toga proces 1 prebrise tu poruku svojom novom porukom. Nakon toga, proces 2 cita poruku iz
    // sanduceta i ispisuje je na ekran.
    procesi[0].operacije = {Operacija(POSALJI, "Stara_poruka"), PREPUSTI};
    procesi[0].pauza = 0;
    procesi[1].operacije = {Operacija(POSALJI, "Nova_poruka"), PREPUSTI};
    procesi[1].pauza = 100;
    procesi[2].operacije = {PRIMI};
    procesi[2].pauza = 200;

    izvrsiProcese(os, procesi);
    cout << endl;
}

void testRazmenaPorukeIzmedjuDvaProcesaSaCekanjem() {
    cout << "TEST: razmena poruke izmedju 2 procesa, proces koji prima poruku mora da ceka na "
         << "poruku" << endl;
    cout << "********************" << endl;
    OS os;

    vector<Proces> procesi(2);

    // Razmena poruke izmedju dva procesa. Proces 0 odmah pokusa da primi poruku, sto nije moguce
    // dok je sanduce prazno, pa on mora da ide u cekanje. Nakon toga, proces 1 postavlja poruku u
    // sanduce, sto treba da probudi proces 0.
    procesi[0].operacije = {PRIMI, NORMALNA};
    procesi[0].pauza = 0;
    procesi[1].operacije = {Operacija(POSALJI, "Zdravo"), PREPUSTI};
    procesi[1].pauza = 100;

    izvrsiProcese(os, procesi);
    cout << endl;
}

void testRazmenaPorukeIzmedjuTriProcesaSaCekanjem() {
    cout << "TEST: razmena poruke izmedju 3 procesa, proces koji prima poruku mora da ceka na "
         << "poruku" << endl;
    cout << "********************" << endl;
    OS os;

    vector<Proces> procesi(3);

    // Test scenario 3: razmena poruke izmedju dva procesa
    // Proces 0 salje poruku "Zdravo", Proces 1 prima tu poruku i ispisuje je na ekran.
    procesi[0].operacije = {PRIMI, NORMALNA};
    procesi[0].pauza = 0;
    procesi[1].operacije = {PRIMI, NORMALNA};
    procesi[1].pauza = 100;
    procesi[2].operacije = {Operacija(POSALJI, "Vise_primalaca"), PREPUSTI};
    procesi[2].pauza = 200;

    izvrsiProcese(os, procesi);
    cout << endl;
}

int main() {
    // Na ovom mestu se izvrsavaju sva definisana test scenarija. Na pocetku su svi testovi, sem prvog, zakomentarisani.
    // Kako budete implementirali delove zadatka, mozete da otkomentarisete i ostale testove:

    testPrekljuciNaKrajuIzvrsenja();
    //testPrekljuciNaOperacijiPrepusti();
    //testRazmenaPorukeIzmedjuDvaProcesaBezCekanja();
    //testRazmenaPorukeIzmedjuTriProcesaBezCekanja();
    //testRazmenaPorukeIzmedjuDvaProcesaSaCekanjem();
    //testRazmenaPorukeIzmedjuTriProcesaSaCekanjem();
}
